/*	$Id: cfg.c,v 1.25 2015/06/06 21:00:42 kristaps Exp $ */
/*
 * Copyright (c) 2014, 2015 Kristaps Dzonsons <kristaps@kcons.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <gd.h>
#include <gsl/gsl_rng.h>
#include <SDL_mixer.h>

#include <flocks.h>
#include "extern.h"

const char *cfgkeys[CFGKEY__MAX] = {
	"align", /* CFGKEY_ALIGN */
	"alignflock", /* CFGKEY_ALIGNFLOCK */
	"birdsperflock", /* CFGKEY_BIRDSPERFLOCK */
	"bps", /* CFGKEY_BPS */
	"cohesion", /* CFGKEY_COHESION */
	"dimensions", /* CFGKEY_DIMENSIONS */
	"ground", /* CFGKEY_GROUND */
	"palign", /* CFGKEY_PALIGN */
	"palignflock", /* CFGKEY_PALIGNFLOCK */
	"payoffs", /* CFGKEY_PAYOFFS */
	"playinterval", /* CFGKEY_PLAYINTERVAL */
	"roosts", /* CFGKEY_ROOSTS */
	"scattermean", /* CFGKEY_SCATTERMEAN */
	"separation", /* CFGKEY_SEPARATION */
	"speedmax", /* CFGKEY_SPEED_MAX */
	"speedmin", /* CFGKEY_SPEED_MIN */
	"tendrdest", /* CFGKEY_TEND_RDEST */
	"tendroost", /* CFGKEY_TEND_ROOST */
	"tendfromroost", /* CFGKEY_TEND_FROM_ROOST */
	"window", /* CFGKEY_WINDOW */
};

const char *fecfgkeys[FECFGKEY__MAX] = {
	"autotrack", /* FECFGKEY_AUTOTRACK */
	"fps", /* FECFGKEY_FPS */
	"hudbig", /* FECFGKEY_HUDBIG */
	"hudlastplay", /* FECFGKEY_HUDLASTPLAY */
	"hudmini", /* FECFGKEY_HUDMINI */
	"lightpos", /* FECFGKEY_LIGHTPOS */
	"maxframes", /* FECFGKEY_MAXFRAMES */
	"maxmatches", /* FECFGKEY_MAXMATCHES */
	"noise0", /* FECFGKEY_NOISE0 */
	"objects", /* FECFGKEY_OBJECTS */
	"output", /* FECFGKEY_OUTPUT */
	"res", /* FECFGKEY_RES */
	"scale", /* FECFGKEY_SCALE */
	"textures", /* FECFGKEY_TEXTURES */
	"window", /* FECFGKEY_WINDOW */
};

static int
cfg_align(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->align = v;
	return(1);
}

static int
cfg_alignflock(struct game *g, double v)
{

	if (v < 0.0)
		v = 0.0;
	if (v > 1.0)
		v = 1.0;
	g->alignflock = v;
	return(1);
}

static int
cfg_birdsperflock(struct game *g, int v)
{

	if (v < 1)
		return(0);
	g->sz = v;
	return(1);
}


static int
cfg_bps(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->bps = v;
	return(1);
}

static int
cfg_cohesion(struct game *g, double v)
{

	if (v < 0.0 || v > 1.0)
		return(0);
	g->cohesion = v;
	return(1);
}

static int
cfg_dimensions(struct game *g, double v)
{

	if (v < 1.0)
		return(0);
	g->dims = v;
	return(1);
}

static int
cfg_ground(struct game *g, double v)
{

	if (v <= 0.0 || v >= 1.0)
		return(0);
	g->ground = v;
	return(1);
}

static int
cfg_palign(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->palign = v;
	return(1);
}

static int
cfg_palignflock(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->palignflock = v;
	return(1);
}

static int
cfg_payoffs(struct game *g, const double *poffs, size_t strats)
{
	size_t	i, j, k;

	for (i = 0; i < strats * strats; i++)
		if (poffs[i] > 1.0 || poffs[i] < 0.0)
			return(0);

	assert(NULL == g->game);
	g->game = calloc(strats, sizeof(double *));
	g->flocks = strats;
	for (i = k = 0; i < strats; i++) {
		g->game[i] = calloc(strats, sizeof(double));
		for (j = 0; j < strats; j++, k++)
			g->game[i][j] = poffs[k];
	}

	return(1);
}

static int
cfg_playinterval(struct game *g, double v)
{

	if (v <= 0.0)
		return(0);
	g->play_interval = v;
	return(1);
}

static int
cfg_roosts(struct game *g, int v)
{

	if (v < 1)
		return(0);
	g->roosts = (size_t)v;
	return(1);
}

static int
cfg_scattermean(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->scattermean = v;
	return(1);
}

static int
cfg_separation(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->separation = v;
	return(1);
}

static int
cfg_speed_max(struct game *g, double v)
{

	if (v <= 0.0 || v <= g->minspeed)
		return(0);
	g->maxspeed = v;
	return(1);
}

static int
cfg_speed_min(struct game *g, double v)
{

	if (v <= 0.0 || v >= g->maxspeed)
		return(0);
	g->minspeed = v;
	return(1);
}

static int
cfg_tendrdest(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->tend_to_rdest = v;
	return(1);
}

static int
cfg_tendfromroost(struct game *g, double v)
{

	if (v > 0.0)
		return(0);
	g->tend_from_roost = v;
	return(1);
}

static int
cfg_tendroost(struct game *g, double v)
{

	if (v < 0.0)
		return(0);
	g->tend_to_roost = v;
	return(1);
}

static int
cfg_window(struct game *g, int v)
{

	if (v <= 0)
		return(0);
	g->windowsz = v;
	return(1);
}

static int
cfgkey_align(struct game *g, const char *cp)
{

	return(cfg_align(g, NULL == cp ? 0.018 : atof(cp)));
}

static int
cfgkey_alignflock(struct game *g, const char *cp)
{

	return(cfg_alignflock(g, NULL == cp ? 0.0 : atof(cp)));
}

static int
cfgkey_cohesion(struct game *g, const char *cp)
{

	return(cfg_cohesion(g, NULL == cp ? 0.15 : atof(cp)));
}

static int
cfgkey_birdsperflock(struct game *g, const char *cp)
{

	return(cfg_birdsperflock(g, NULL == cp ? 50 : atoi(cp)));
}

static int
cfgkey_bps(struct game *g, const char *cp)
{

	return(cfg_bps(g, NULL == cp ? 5.0 : atof(cp)));
}

static int
cfgkey_dimensions(struct game *g, const char *cp)
{

	return(cfg_dimensions(g, NULL == cp ? 100.0 : atof(cp)));
}

static int
cfgkey_ground(struct game *g, const char *cp)
{

	return(cfg_ground(g, NULL == cp ? 0.5 : atof(cp)));
}

static int
cfgkey_palign(struct game *g, const char *cp)
{

	return(cfg_palign(g, NULL == cp ? 40.0 : atof(cp)));
}

static int
cfgkey_palignflock(struct game *g, const char *cp)
{

	return(cfg_palignflock(g, NULL == cp ? 30.0 : atof(cp)));
}

static int
cfgkey_payoffs(struct game *g, const char *cp)
{
	double	 poffs[49];
	char	*cpp, *sv;
	size_t	 i, tokens, strats;

	if (NULL == cp) {
		poffs[0] = 0.5;
		poffs[1] = 0.2;
		poffs[2] = 0.7;
		poffs[3] = 0.3;
		return(cfg_payoffs(g, poffs, 2));
	}

	cpp = sv = strdup(cp);
	tokens = 1;

	while (NULL != (cpp = strchr(cpp, ','))) {
		tokens++;
		cpp++;
	}
	
	switch (tokens) {
	case (4):
		strats = 2;
		break;
	case (9):
		strats = 3;
		break;
	case (16):
		strats = 4;
		break;
	case (25):
		strats = 5;
		break;
	case (36):
		strats = 6;
		break;
	case (49):
		strats = 7;
		break;
	default:
		free(sv);
		return(0);
	}

	for (i = 0; i < tokens; i++) {
		if (NULL != (cpp = strchr(cp, ',')))
			*cpp++ = '\0';
		poffs[i] = atof(cp);
		cp = cpp;
	}

	free(sv);
	return(cfg_payoffs(g, poffs, strats));
}

static int
cfgkey_playinterval(struct game *g, const char *cp)
{

	return(cfg_playinterval(g, NULL == cp ? 0.5 : atof(cp)));
}

static int
cfgkey_roosts(struct game *g, const char *cp)
{

	return(cfg_roosts(g, NULL == cp ? 1 : atoi(cp)));
}

static int
cfgkey_scattermean(struct game *g, const char *cp)
{

	return(cfg_scattermean(g, NULL == cp ? 4.0 : atof(cp)));
}

static int
cfgkey_separation(struct game *g, const char *cp)
{

	return(cfg_separation(g, NULL == cp ? 2.0 : atof(cp)));
}

static int
cfgkey_speed_max(struct game *g, const char *cp)
{

	return(cfg_speed_max(g, NULL == cp ? 15.0 : atof(cp)));
}

static int
cfgkey_speed_min(struct game *g, const char *cp)
{

	return(cfg_speed_min(g, NULL == cp ? 7.5 : atof(cp)));
}

static int
cfgkey_tendrdest(struct game *g, const char *cp)
{

	return(cfg_tendrdest(g, NULL == cp ? 0.005 : atof(cp)));
}

static int
cfgkey_tendfromroost(struct game *g, const char *cp)
{

	return(cfg_tendfromroost(g, NULL == cp ? -0.001 : atof(cp)));
}

static int
cfgkey_window(struct game *g, const char *cp)
{

	return(cfg_window(g, NULL == cp ? 100 : atoi(cp)));
}

static int
cfgkey_tendroost(struct game *g, const char *cp)
{

	return(cfg_tendroost(g, NULL == cp ? 0.0 : atof(cp)));
}

int
cfg_scalar_int(struct game *g, enum cfgkey key, int v)
{

	switch (key) {
	case (CFGKEY_BIRDSPERFLOCK):
		return(cfg_birdsperflock(g, v));
	case (CFGKEY_ROOSTS):
		return(cfg_roosts(g, v));
	case (CFGKEY_WINDOW):
		return(cfg_window(g, v));
	default:
		abort();
	}
	return(0);
}

int
cfg_scalar_double(struct game *g, enum cfgkey key, double v)
{

	switch (key) {
	case (CFGKEY_ALIGN):
		return(cfg_align(g, v));
	case (CFGKEY_ALIGNFLOCK):
		return(cfg_alignflock(g, v));
	case (CFGKEY_BPS):
		return(cfg_bps(g, v));
	case (CFGKEY_COHESION):
		return(cfg_cohesion(g, v));
	case (CFGKEY_DIMENSIONS):
		return(cfg_dimensions(g, v));
	case (CFGKEY_GROUND):
		return(cfg_ground(g, v));
	case (CFGKEY_PALIGN):
		return(cfg_palign(g, v));
	case (CFGKEY_PALIGNFLOCK):
		return(cfg_palignflock(g, v));
	case (CFGKEY_PLAYINTERVAL):
		return(cfg_playinterval(g, v));
	case (CFGKEY_SCATTERMEAN):
		return(cfg_scattermean(g, v));
	case (CFGKEY_SEPARATION):
		return(cfg_separation(g, v));
	case (CFGKEY_SPEED_MAX):
		return(cfg_speed_max(g, v));
	case (CFGKEY_SPEED_MIN):
		return(cfg_speed_min(g, v));
	case (CFGKEY_TEND_RDEST):
		return(cfg_tendrdest(g, v));
	case (CFGKEY_TEND_ROOST):
		return(cfg_tendroost(g, v));
	case (CFGKEY_TEND_FROM_ROOST):
		return(cfg_tendfromroost(g, v));
	default:
		abort();
	}
	return(0);
}

int
cfg_string(struct game *g, enum cfgkey key, const char *v)
{

	switch (key) {
	case (CFGKEY_ALIGN):
		return(cfgkey_align(g, v));
	case (CFGKEY_ALIGNFLOCK):
		return(cfgkey_alignflock(g, v));
	case (CFGKEY_BIRDSPERFLOCK):
		return(cfgkey_birdsperflock(g, v));
	case (CFGKEY_BPS):
		return(cfgkey_bps(g, v));
	case (CFGKEY_COHESION):
		return(cfgkey_cohesion(g, v));
	case (CFGKEY_DIMENSIONS):
		return(cfgkey_dimensions(g, v));
	case (CFGKEY_GROUND):
		return(cfgkey_ground(g, v));
	case (CFGKEY_PALIGN):
		return(cfgkey_palign(g, v));
	case (CFGKEY_PALIGNFLOCK):
		return(cfgkey_palignflock(g, v));
	case (CFGKEY_PLAYINTERVAL):
		return(cfgkey_playinterval(g, v));
	case (CFGKEY_PAYOFFS):
		return(cfgkey_payoffs(g, v));
	case (CFGKEY_ROOSTS):
		return(cfgkey_roosts(g, v));
	case (CFGKEY_SCATTERMEAN):
		return(cfgkey_scattermean(g, v));
	case (CFGKEY_SEPARATION):
		return(cfgkey_separation(g, v));
	case (CFGKEY_SPEED_MAX):
		return(cfgkey_speed_max(g, v));
	case (CFGKEY_SPEED_MIN):
		return(cfgkey_speed_min(g, v));
	case (CFGKEY_TEND_RDEST):
		return(cfgkey_tendrdest(g, v));
	case (CFGKEY_TEND_ROOST):
		return(cfgkey_tendroost(g, v));
	case (CFGKEY_TEND_FROM_ROOST):
		return(cfgkey_tendfromroost(g, v));
	case (CFGKEY_WINDOW):
		return(cfgkey_window(g, v));
	default:
		abort();
	}

	return(0);
}

/*
 * Check if all configuration keys (-Ofoo=bar) are valid.
 */
int
cfg_check(struct game *g, const char **cfg)
{
	enum cfgkey	 key;

	for (key = 0; key < CFGKEY__MAX; key++) {
		if (cfg_string(g, key, cfg[key]))
			continue;
		fprintf(stderr, "%s: bad config\n", cfgkeys[key]);
		return(0);
	}
	return(1);
}

int
cfg_parsekey(char *cp, char **cfg, size_t keysz, const char *const *keys)
{
	char	*key, *val;
	size_t	 i;

	key = cp;
	if (NULL != (val = strchr(cp, '=')))
		*val++ = '\0';
	else
		val = cp + strlen(cp);

	for (i = 0; i < keysz; i++)
		if (0 == strcmp(key, keys[i]))
			break;

	if (i == keysz) {
		fprintf(stderr, "%s: no such configuration\n", key);
		return(0);
	}

	cfg[i] = val;
	return(1);
}
