/*	$Id: display.c,v 1.22 2015/06/06 21:00:42 kristaps Exp $ */
/*
 * Copyright (c) 2014, 2015 Kristaps Dzonsons <kristaps@kcons.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __linux__
#include <bsd/string.h>
#endif
#include <unistd.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <gd.h>
#include <gsl/gsl_rng.h>
#include <flocks.h>
#include <SDL.h>
#include <SDL_mixer.h>

#include "extern.h"

/*
 * Flock colours.
 * These are stolen from the bmigrate utility, which in turn was
 * inspired by gnuplot(1)'s non-default palette.
 */
#define	SIZE_COLOURS 4
static	const float colours[SIZE_COLOURS][3] = {
	{ 255 / 255.0, 219 / 255.0, 88 / 255.0 }, /* Mustard (yellow) */
	{ 128 / 255.0, 0 / 255.0, 0 / 255.0 }, /* Maroon (red) */
	{ 42 / 255.0, 82 / 255.0, 190 / 255.0 }, /* Cerulean blue (blue) */
	{ 150 / 255.0, 113 / 255.0, 23 / 255.0 }, /* Sandy Tuap (yellow) */
};

static	struct args	   g_arg;

static void
flocks_record_stats(void)
{
	struct game	*g = g_arg.g;
	size_t		 i;

	printf("%zu", g->matches);

	for (i = 0; i < g->flocks; i++) {
		printf(" %zu", g->flocksz[i]);
		printf(" %zu", g->windowhits[i]);
		printf(" %zu", g->windowmatches[i]);
	}

	putchar('\n');
	fflush(stdout);
}

static int
flocks_check_exit(void)
{

	if (g_arg.frame >= g_arg.maxframes)
		return(1);
	if (g_arg.g->matches >= g_arg.maxmatches)
		return(1);

	return(0);
}

/*
 * Update the camera to either point in a user-defined direction (i.e.,
 * the view angles specified by the UI) or auto-track the center of
 * mass.
 */
static void
flocks_update_camera(double elapsed)
{
	double	cx, cy, cz, delta, r, new, cvxz, cvxy;

	if (g_arg.pause)
		return;

	/* 
	 * This is the minimum we can travel if we were to start
	 * decelerating at this exact moment.
	 * It's just the integral of the linear decreasing slope.
	 */
	cvxz = (g_arg.cvxz * (M_PI / 360.0)) * 
		g_arg.cvxz * 0.5 * elapsed;
	if (g_arg.cvxz < 0.0)
		cvxz = -cvxz;

	cvxy = (g_arg.cvxy * (M_PI / 360.0)) * 
		g_arg.cvxy * 0.5 * elapsed;
	if (g_arg.cvxy < 0.0)
		cvxy = -cvxy;

	if (g_arg.autotrack) {
		/* Distance in all three directions. */
		cx = g_arg.g->allpossum.x;
		cy = g_arg.g->allpossum.y;
		cz = g_arg.g->allpossum.z;

		/* Get the xy angle of the new centre. */
		if ((new = atan2(cx, cz)) < 0.0)
			new += M_PI * 2.0;
		else if (new > M_PI * 2.0)
			new -= M_PI * 2.0;

		/* Difference between us and centre. */
		delta = atan2
			(sin(new - g_arg.cxz), 
			 cos(new - g_arg.cxz));

		/* 
		 * Do we adjust our velocity? 
		 * Cap at 40 increments.
		 */
		if (delta > 0.0 && delta > cvxz) {
			g_arg.cvxz += 1.0;
			if (g_arg.cvxz > 40.0)
				g_arg.cvxz = 40.0;
		} else if (delta > 0.0) {
			g_arg.cvxz -= 1.0;
			if (g_arg.cvxz < 0.0)
				g_arg.cvxz = 0.0;
		} else if (delta < 0.0 && delta < cvxz) {
			g_arg.cvxz -= 1.0;
			if (g_arg.cvxz < -40.0)
				g_arg.cvxz = -40.0;
		} else if (delta < 0.0) {
			g_arg.cvxz += 1.0;
			if (g_arg.cvxz > 0.0)
				g_arg.cvxz = 0.0;
		}

		/* Get the vertical angle of the new centre. */
		r = sqrt(cx * cx + cz * cz);
		new = atan2(cy, r);

		if (new > M_PI_2)
			new = M_PI_2;
		else if (new < 0.0)
			new = 0.0;

		/* Difference between us and centre. */
		delta = atan2
			(sin(new - g_arg.cxy), 
			 cos(new - g_arg.cxy));

		/* 
		 * Do we adjust our velocity? 
		 * Cap at 40 increments.
		 */
		if (delta > 0.0 && delta > cvxy) {
			g_arg.cvxy += 1.0;
			if (g_arg.cvxy > 40.0)
				g_arg.cvxy = 40.0;
		} else if (delta > 0.0) {
			g_arg.cvxy -= 1.0;
			if (g_arg.cvxy < 0.0)
				g_arg.cvxy = 0.0;
		} else if (delta < 0.0 && delta < cvxy) {
			g_arg.cvxy -= 1.0;
			if (g_arg.cvxy < -40.0)
				g_arg.cvxy = -40.0;
		} else if (delta < 0.0) {
			g_arg.cvxy += 1.0;
			if (g_arg.cvxy > 0.0)
				g_arg.cvxy = 0.0;
		}
	} 

	/* Are we moving?  Adjust our angle position. */
	g_arg.cxz += g_arg.cvxz * (M_PI / 360.0) * elapsed;
	g_arg.cxy += g_arg.cvxy * (M_PI / 360.0) * elapsed;

	/* Adjust angle to 2PI space. */
	if (g_arg.cxz < 0.0)
		g_arg.cxz += M_PI * 2.0;
	else if (g_arg.cxz > M_PI * 2.0)
		g_arg.cxz -= M_PI * 2.0;
	if (g_arg.cxy < 0.0)
		g_arg.cxy += M_PI * 2.0;
	else if (g_arg.cxy > M_PI * 2.0)
		g_arg.cxy -= M_PI * 2.0;

	g_arg.cx = sin(g_arg.cxz);
	g_arg.cy = sin(g_arg.cxy);
	g_arg.cz = cos(g_arg.cxz);
}

/*
 * Update display mode.
 * Groks the elapsed time (in microseconds) from the last display and
 * stipulates that the birds move that distance/time.
 */
static void
flocks_update_timeo(int value)
{
	struct timeval	tv, diff;
	size_t		sv;
	double		v, cx, cy, cz, d;

	memset(&tv, 0, sizeof(struct timeval));
	memset(&diff, 0, sizeof(struct timeval));
	(void)gettimeofday(&tv, NULL);

	v = 0.0;
	if (timerisset(&g_arg.tv)) {
		timersub(&tv, &g_arg.tv, &diff);
		v = (double)diff.tv_sec + (diff.tv_usec / 1000000.0);
		g_arg.fps_real = 1.0 / v;
	} 

	if ( ! g_arg.pause) {
		sv = g_arg.g->matches;
		birds_update(g_arg.g, v);
		if (sv != g_arg.g->matches) {
			flocks_record_stats();
		}
	}

	if (NULL != g_arg.noise) {
		cx = g_arg.g->allpossum.x / 
			(double)g_arg.g->sz / g_arg.g->dims;
		cy = g_arg.g->allpossum.y / 
			(double)g_arg.g->sz / g_arg.g->dims;
		cz = g_arg.g->allpossum.z / 
			(double)g_arg.g->sz / g_arg.g->dims;
		d = 1.0 + 3.0 * sqrt(cx * cx + cy * cy + cz * cz);
		assert(g_arg.noisechan >= 0);
		Mix_Volume(g_arg.noisechan, MIX_MAX_VOLUME * 1.0 / (d * d));
	}

	flocks_update_camera(v);

	memcpy(&g_arg.tv, &tv, sizeof(struct timeval));
	glutPostRedisplay();
	glutTimerFunc(g_arg.refresh, flocks_update_timeo, -1);
}

/*
 * Update record mode.
 * Here our frames-per-second are fixed, so move the birds in the fixed
 * amount of time.
 */
static void
flocks_update_idle(void)
{
	struct timeval	 tv, diff;
	size_t		 sv;
	double		 v;

	memset(&tv, 0, sizeof(struct timeval));
	(void)gettimeofday(&tv, NULL);

	v = 0.0;
	if (timerisset(&g_arg.tv)) {
		timersub(&tv, &g_arg.tv, &diff);
		v = (double)diff.tv_sec + (diff.tv_usec / 1000000.0);
		g_arg.fps_real = 1.0 / v;
	} 

	sv = g_arg.g->matches;
	birds_update(g_arg.g, 1.0 / (double)g_arg.fps);
	if (sv != g_arg.g->matches)
		flocks_record_stats();

	flocks_update_camera(1.0 / (double)g_arg.fps);

	memcpy(&g_arg.tv, &tv, sizeof(struct timeval));
	glutPostRedisplay();
	glutIdleFunc(flocks_update_idle);
}

/*
 * Draw an individual bird.
 * The bird should be oriented toward the direction of travel (i.e.,
 * toward the velocity vector).
 */
static void
flocks_bird_draw(const struct bird *b)
{
	struct vector	cnorm, dnorm, raxis;
	double		rrot;
	size_t		wingpos;

	wingpos = lround(b->wingpos * (WING_SIZE - 1));
	assert(wingpos < WING_SIZE);
	vector_zero(&cnorm);
	cnorm.x = 1.0;
	dnorm = b->velocity;
	vector_norm(&dnorm);
	vector_crossproduct(&cnorm, &dnorm, &raxis);
	vector_norm(&raxis);
	rrot = acos(vector_dotproduct(&cnorm, &dnorm));

	glRotated(rrot * 180.0 / M_PI, raxis.x, raxis.y, raxis.z);
	glVertexPointer(3, GL_FLOAT, 0, g_arg.birdVerts[wingpos]);
	glNormalPointer(GL_FLOAT, 0, g_arg.birdNormals[wingpos]);
	glDrawArrays(GL_TRIANGLES, 0, g_arg.birdNumVerts[wingpos]);
	assert(glGetError() == GL_NO_ERROR);
}

/*
 * Paint each bird on the screen.
 * Also note whether it's been "lit up" (i.e., has played a game and
 * changed strategy) by colouring it red.
 */
static void
flocks_update_bird(const struct bird *b)
{
	GLfloat colour[4];

	glPushMatrix();
	glTranslated(b->position.x, b->position.y, b->position.z);
	if (1.0 != g_arg.scale)
		glScaled(g_arg.scale, g_arg.scale, g_arg.scale);
	colour[0] = 0.0;
	colour[1] = 0.0;
	colour[2] = 0.0;
	colour[3] = 1.0;

	if (b->removed) {
		glMaterialfv(GL_FRONT, GL_SPECULAR, colour);
		colour[0] = 0.5;
		colour[1] = 0.5;
		colour[2] = 0.5;
		colour[3] = 1.0;
	} else if (b->litup > 0.0) {
		colour[0] = 0.0;
		colour[1] = 0.0;
		colour[2] = 1.0;
		colour[3] = 1.0;
		glMaterialfv(GL_FRONT, GL_SPECULAR, colour);
		colour[3] = 1.0;
	} else if (b->newborn < 1.0) {
		glMaterialfv(GL_FRONT, GL_SPECULAR, colour);
		colour[0] = colours[b->flock % SIZE_COLOURS][0];
		colour[1] = colours[b->flock % SIZE_COLOURS][1];
		colour[2] = colours[b->flock % SIZE_COLOURS][2];
		colour[3] = b->newborn;
	} else {
		glMaterialfv(GL_FRONT, GL_SPECULAR, colour);
		colour[0] = colours[b->flock % SIZE_COLOURS][0];
		colour[1] = colours[b->flock % SIZE_COLOURS][1];
		colour[2] = colours[b->flock % SIZE_COLOURS][2];
		colour[3] = 1.0;
	}

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colour);
	assert(glGetError() == GL_NO_ERROR);

	if (g_arg.drawaxes || g_arg.drawnoobjs)
		glutSolidSphere(g_arg.g->separation / 3.0, 20, 20);
	else
		flocks_bird_draw(b);

	glPopMatrix();
}

static void 
drawSpace(void)
{
	GLfloat		 colour[4];

	glBegin(GL_LINES);
	colour[0] = 1.0;
	colour[1] = 0.0;
	colour[2] = 0.0;
	colour[3] = 1.0;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour);
	glVertex3d(-g_arg.g->dims, 0.0, 0.0);
	glVertex3d(g_arg.g->dims, 0.0, 0.0);
	colour[0] = 0.0;
	colour[1] = 1.0;
	colour[2] = 0.0;
	colour[3] = 1.0;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, colour);
	glVertex3d(0.0, -g_arg.g->dims, 0.0);
	glVertex3d(0.0, g_arg.g->dims, 0.0);
	glEnd();
}

/*
 * Draw our "skybox" cube of textures.
 * This is enacted by taking each "skybox" texture and putting it on the
 * corresponding face of a cube which is at the maximum boundary of our
 * space.
 */
static void 
drawBackground(void)
{
	double	minx, maxx, miny, maxy, minz, maxz;

	/*
	 * We don't want lighting or culling.
	 * The latter will paint on the cube, which we don't want (it's
	 * already painted!) and as we can't be bothered with winding,
	 * disregard the former.
	 */
	glDepthMask(GL_FALSE);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);

	minx = -g_arg.g->dims;
	miny = -g_arg.g->dims;
	minz = -g_arg.g->dims;
	maxx = g_arg.g->dims;
	maxy = g_arg.g->dims;
	maxz = g_arg.g->dims;

	glColor3d(1.0, 1.0, 1.0);

	if (g_arg.bgid[FACE_NEGZ]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_NEGZ]);
		glBegin(GL_QUADS);
		glTexCoord2d(0, 0);
		glVertex3d(minx, miny, minz);
		glTexCoord2d(0.0, 1.0);
		glVertex3d(minx, maxy, minz);
		glTexCoord2d(1, 1);
		glVertex3d(maxx, maxy, minz);
		glTexCoord2d(1, 0);
		glVertex3d(maxx, miny, minz);
		glEnd();
	}

	if (g_arg.bgid[FACE_POSY]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_POSY]);
		glBegin(GL_QUADS);
		glTexCoord2f(1, 1);
		glVertex3d(minx, maxy, minz);
		glTexCoord2f(0, 1);
		glVertex3d(maxx, maxy, minz);
		glTexCoord2f(0, 0);
		glVertex3d(maxx, maxy, maxz);
		glTexCoord2f(1, 0);
		glVertex3d(minx, maxy, maxz);
		glEnd();
	}

	if (g_arg.bgid[FACE_POSX]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_POSX]);
		glBegin(GL_QUADS);
		glTexCoord2f(1, 0);
		glVertex3d(minx, miny, minz);
		glTexCoord2f(0, 0);
		glVertex3d(minx, miny, maxz);
		glTexCoord2f(0, 1);
		glVertex3d(minx, maxy, maxz);
		glTexCoord2f(1, 1);
		glVertex3d(minx, maxy, minz);
		glEnd();
	}

	if (g_arg.bgid[FACE_NEGX]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_NEGX]);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex3d(maxx, miny, minz);
		glTexCoord2f(1, 0);
		glVertex3d(maxx, miny, maxz);
		glTexCoord2f(1, 1);
		glVertex3d(maxx, maxy, maxz);
		glTexCoord2f(0, 1);
		glVertex3d(maxx, maxy, minz);
		glEnd();
	}

	if (g_arg.bgid[FACE_POSZ]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_POSZ]);
		glBegin(GL_QUADS);
		glTexCoord2f(1, 0);
		glVertex3d(minx, miny, maxz);
		glTexCoord2f(1, 1);
		glVertex3d(minx, maxy, maxz);
		glTexCoord2f(0, 1);
		glVertex3d(maxx, maxy, maxz);
		glTexCoord2f(0, 0);
		glVertex3d(maxx, miny, maxz);
		glEnd();
	}

	if (g_arg.bgid[FACE_NEGY]) {
		glBindTexture(GL_TEXTURE_2D, g_arg.bgid[FACE_NEGY]);
		glBegin(GL_QUADS);
		glTexCoord2f(1, 0);
		glVertex3d(minx, miny, minz);
		glTexCoord2f(1, 1);
		glVertex3d(minx, miny, maxz);
		glTexCoord2f(0, 1);
		glVertex3d(maxx, miny, maxz);
		glTexCoord2f(0, 0);
		glVertex3d(maxx, miny, minz);
		glEnd();
	}

	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glDepthMask(GL_TRUE);
}

static void 
flocks_display(void)
{
	double	 maxdim;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	/* The maximum distance in 3-space. */
	maxdim = sqrt(sqrt(2.0 * g_arg.g->dims * g_arg.g->dims) *
		      sqrt(2.0 * g_arg.g->dims * g_arg.g->dims) + 
		      g_arg.g->dims * 
		      g_arg.g->dims);
	gluPerspective(g_arg.cradius, 
		g_arg.width / (double)g_arg.height, 
		1.0, maxdim + 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	gluLookAt(0.0, 0.0, 0.0, 
		g_arg.cx, g_arg.cy, g_arg.cz, 
		0.0, 1.0, 0.0);

	glEnable(GL_BLEND); 
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (g_arg.drawaxes || g_arg.drawnotexs)
		drawSpace();
	else
		drawBackground();

	glEnable(GL_RESCALE_NORMAL);
	birds_iterate(g_arg.g, flocks_update_bird);
	glDisable(GL_RESCALE_NORMAL);
	glDisable(GL_BLEND);

	if (g_arg.drawhud)
		hudDraw(&g_arg);

	glutSwapBuffers();

	++g_arg.frame;
	if (flocks_check_exit())
		exit(EXIT_SUCCESS);

	if (NULL == g_arg.ffmpeg)
		return;
	
	/*
	 * Extract the pixel buffer into our own buffer.
	 * Then use GD to format it as a JPG and push it out into a
	 * file, which we'll later render with ffmpeg.
	 */
	glReadPixels(0, 0, g_arg.width, 
		g_arg.height, GL_RGBA, 
		GL_UNSIGNED_BYTE, g_arg.data);
	fwrite(g_arg.data, 4 * g_arg.width * 
		g_arg.height, 1, g_arg.ffmpeg);
}

static void
flocks_special(int key, int x, int y)
{

	if (HUDVIEW_BIG == g_arg.drawhud) {
		hudSpecial(&g_arg, key, x, y);
		glutPostRedisplay();
		return;
	} else if (g_arg.autotrack)
		return;

	switch (key) {
	case (GLUT_KEY_DOWN):
		g_arg.cxy -= M_PI / 180.0;
		if (g_arg.cxy < 0.0)
			g_arg.cxy = 0.0;
		break;
	case (GLUT_KEY_UP):
		g_arg.cxy += M_PI / 180.0;
		if (g_arg.cxy > M_PI_4)
			g_arg.cxy = M_PI_4;
		break;
	case (GLUT_KEY_LEFT):
		g_arg.cxz += M_PI / 180.0;
		break;
	case (GLUT_KEY_RIGHT):
		g_arg.cxz -= M_PI / 180.0;
		break;
	default:
		break;
	}
}

static void
flocks_mouse(int button, int state, int x, int y)
{

	if (g_arg.drawhud) {
		hudMouse(&g_arg, button, state, x, g_arg.height - y);
		glutPostRedisplay();
	}
}

/*
 * Handle keypresses.
 * Just do a few of them; add as needed.
 */
static void
flocks_keyboard(unsigned char key, int x, int y)
{
	double	 h;

	switch (key) {
	case ('a'):
		g_arg.autotrack = !g_arg.autotrack;
		if ( ! g_arg.autotrack)
			g_arg.cvxz = g_arg.cvxy = 0.0;
		break;
	case ('s'):
		if (g_arg.g->scattering <= 0.0)
			g_arg.g->nextscattering = FLT_EPSILON;
		break;
	case ('b'):
		g_arg.drawaxes = !g_arg.drawaxes;
		break;
	case ('f'):
		if (g_arg.fullscreen) {
			glutReshapeWindow(g_arg.widthsv, g_arg.heightsv);
			g_arg.fullscreen = 0;
		} else {
			g_arg.widthsv = g_arg.width;
			g_arg.heightsv =  g_arg.height;
			g_arg.fullscreen = 1;
			glutFullScreen();
		}
		break;
	case (27): /* Escape key */
		g_arg.drawhud = 0;
		break;
	case ('h'):
		g_arg.drawhud = HUDVIEW_BIG;
		break;
	case ('P'):
		g_arg.drawhud = HUDVIEW_LASTPLAY;
		break;
	case ('H'):
		g_arg.drawhud = HUDVIEW_MINI;
		break;
	case ('l'):
		h = 2.0 *
		    sqrt(sqrt(2.0 * g_arg.g->dims * g_arg.g->dims) *
		    sqrt(2.0 * g_arg.g->dims * g_arg.g->dims) + 
		    g_arg.g->dims * 
		    g_arg.g->dims);
		g_arg.lightpos[0] = sin(g_arg.cxz) * h;
		g_arg.lightpos[1] = sin(g_arg.cxy) * h;
		g_arg.lightpos[2] = cos(g_arg.cxz) * h;
		glLightfv(GL_LIGHT0, GL_POSITION, g_arg.lightpos);
		break;
	case ('q'):
		/*
		 * Quit.
		 * Should be handled by WM, but add one in here just to
		 * be certain.
		 */
		exit(EXIT_SUCCESS);
		break;
	case (' '):
	case ('p'):
		/*
		 * Toggle pause.
		 * This simply causes the signal handlers to exit (if
		 * we're not paused) or reinstates them (if we are).
		 */
		if ( ! g_arg.pause) 
			g_arg.pause = 1;
		else
			g_arg.pause = 0;
		if (g_arg.pause && NULL != g_arg.noise) 
			Mix_Pause(g_arg.noisechan);
		else if (NULL != g_arg.noise)
			Mix_Resume(g_arg.noisechan);
		break;
	default:
		break;
	}
}

/*
 * Reshape the window.
 * This is standard OpenGL stuff.
 */
static void 
flocks_reshape(int w, int h)
{

	assert(w > 0);
	assert(h > 0);

	glViewport(0, 0, (GLsizei) w, (GLsizei) h); 

	/* Disallow changes after allocation. */
	if (NULL != g_arg.ffmpeg && 
		(size_t)w != g_arg.width && 
		(size_t)h != g_arg.height) {
		fprintf(stderr, "don't resize and record!\n");
		exit(EXIT_FAILURE);
	}

	g_arg.width = w;
	g_arg.height = h;
}

/*
 * Initialise OpenGL features.
 * This must happen *after* the GLUT window has been created, or else
 * we'll just crash.
 */
static void 
initGL(void)
{

	/*
	 * Set our depth and colour clearing.
	 * Stack depths by less-than-equal.
	 */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glShadeModel(GL_SMOOTH);
	glLightfv(GL_LIGHT0, GL_POSITION, g_arg.lightpos);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	g_arg.cradius = 90.0;
}

static void
onexit(void)
{
	size_t	 i;

	if (NULL != g_arg.ffmpeg)
		pclose(g_arg.ffmpeg);

	for (i = 0; i < WING_SIZE; i++) {
		free(g_arg.birdNormals[i]);
		free(g_arg.birdVerts[i]);
	}

	free(g_arg.filebuf);
	free(g_arg.data);
	birds_free(g_arg.g);
	if (NULL != g_arg.noise)
		Mix_FreeChunk(g_arg.noise);
	Mix_CloseAudio();
	SDL_Quit();
}

/*
 * Run in graphical (GLUT/OpenGL) mode.
 * To do this, we need to instantiate the full GL run-time, which
 * includes the windowing parameters, GLUT itself, etc.
 * Then we perform a single run of the system via glutMainLoop() and the
 * callbacks for idleness or timeout.
 */
int
mainGL(int argc, char *argv[], char *fecfg[], char *cfg[])
{
	const char	*res;
	char		 buf[1024];

	if (-1 == SDL_Init(SDL_INIT_AUDIO)) {
		fprintf(stderr, "audio: %s\n", SDL_GetError());
		return(EXIT_FAILURE);
	} else if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, 
		 MIX_DEFAULT_FORMAT, 2, 1024) < 0) {
		fprintf(stderr, "audio: %s\n", Mix_GetError());
		SDL_Quit();
		return(EXIT_FAILURE);
	} 
	
	if (NULL != fecfg[FECFGKEY_NOISE0]) {
		g_arg.noise = Mix_LoadWAV(fecfg[FECFGKEY_NOISE0]);
		if (NULL == g_arg.noise) 
			fprintf(stderr, "audio: %s\n", Mix_GetError());
	}

	atexit(onexit);

	/* Follow centre of birds. */
	if (NULL != fecfg[FECFGKEY_AUTOTRACK])
		g_arg.autotrack = 1;

	/* Frames per second. */
	g_arg.fps = (NULL != fecfg[FECFGKEY_FPS]) ?
		atoi(fecfg[FECFGKEY_FPS]) : 25;
	if (0 == g_arg.fps) {
		fprintf(stderr, "fps must be >0\n");
		return(EXIT_FAILURE);
	}
	g_arg.refresh = 1.0 / (double)g_arg.fps * 1000.0;

	/* 
	 * Scaling bird wireframe size.
	 * FIXME: this is ugly.
	 */
	g_arg.scale = (NULL != fecfg[FECFGKEY_SCALE]) ?
		atof(fecfg[FECFGKEY_SCALE]) : 1.0;

	/* Termination clauses. */
	g_arg.maxframes = INT_MAX;
	if (NULL != fecfg[FECFGKEY_MAXFRAMES]) 
		g_arg.maxframes = atoi(fecfg[FECFGKEY_MAXFRAMES]);
	g_arg.maxmatches = INT_MAX;
	if (NULL != fecfg[FECFGKEY_MAXMATCHES]) 
		g_arg.maxmatches = atoi(fecfg[FECFGKEY_MAXMATCHES]);

	/* Start with a HUD enabled. */
	if (NULL != fecfg[FECFGKEY_HUDMINI])
		g_arg.drawhud = HUDVIEW_MINI;
	else if (NULL != fecfg[FECFGKEY_HUDBIG])
		g_arg.drawhud = HUDVIEW_BIG;
	else if (NULL != fecfg[FECFGKEY_HUDLASTPLAY])
		g_arg.drawhud = HUDVIEW_LASTPLAY;

	/* 
	 * By default, use a 480p resolution.
	 * Scale up to wide HQ (1440p). 
	 */
	if (NULL == (res = fecfg[FECFGKEY_RES]))
		res = "480p";

	if (0 == strcmp(res, "240p")) {
		g_arg.height = 240;
		g_arg.width = 426;
	} else if (0 == strcmp(res, "360p")) {
		g_arg.height = 360;
		g_arg.width = 640;
	} else if (0 == strcmp(res, "480p")) {
		g_arg.height = 480;
		g_arg.width = 854;
	} else if (0 == strcmp(res, "720p")) {
		g_arg.height = 720;
		g_arg.width = 1280;
	} else if (0 == strcmp(res, "1080p")) {
		g_arg.height = 1080;
		g_arg.width = 1920;
	} else if (0 == strcmp(res, "1440p")) {
		g_arg.height = 1440;
		g_arg.width = 2560;
	} else {
		fprintf(stderr, "%s: bad resolution\n", res);
		return(EXIT_FAILURE);
	}

	/* Sunlight position. */
	g_arg.lightpos[0] = 0.0;
	g_arg.lightpos[1] = 1000.0;
	g_arg.lightpos[2] = 0.0;
	g_arg.lightpos[3] = 1.0;
	if (NULL != fecfg[FECFGKEY_LIGHTPOS])
		if (3 != sscanf(fecfg[FECFGKEY_LIGHTPOS], "%g,%g,%g",
			&g_arg.lightpos[0],
			&g_arg.lightpos[1],
			&g_arg.lightpos[2])) {
			fprintf(stderr, "bad lightpos\n");
			return(EXIT_FAILURE);
		}

	/*
	 * If we're recording, create a buffer for frame filenames.
	 * This prevents us from constantly reallocating in runtime.
	 */
	if (NULL != fecfg[FECFGKEY_OUTPUT]) {
		snprintf(buf, sizeof(buf),
			"ffmpeg -r %zu -f rawvideo -pix_fmt rgba "
			       "-s %zux%zu -i - -threads 0 "
			       "-preset fast -y -pix_fmt yuv420p "
			       "-crf 21 -vf vflip %s",
			       g_arg.fps, g_arg.width, g_arg.height,
			       fecfg[FECFGKEY_OUTPUT]);
		g_arg.ffmpeg = popen(buf, "w");
		if (NULL == g_arg.ffmpeg) {
			perror("ffmpeg");
			return(EXIT_FAILURE);
		}
		g_arg.data = malloc(4 * g_arg.width * g_arg.height);
		assert(NULL != g_arg.data);
	}

	/* 
	 * Load bird wireframe with a simple default.
	 * Failure causes no birds to be drawn.
	 */
	if ( ! objLoadAll(&g_arg, fecfg[FECFGKEY_OBJECTS]))
		g_arg.drawnoobjs = 1;

	/* Initialise the birds structure with some fixed defaults. */
	g_arg.g = birds_alloc
		(2 / (double)(WING_SIZE - 1), 
		 1 / (double)(WING_SIZE - 1));

	/* Read our configuration keys. */
	if ( ! cfg_check(g_arg.g, (const char **)cfg))
		return(EXIT_FAILURE);

	/* Pass the configuration to our birds! */
	birds_init(g_arg.g);

	/*
	 * Initialise our display.
	 * For making videos, we use a single buffer.
	 * For regular display, a double-buffer.
	 * Then install handlers for the keyboard, reshaping the window,
	 * and actually handling display of data.
	 */
	glutInit(&argc, argv);
	glutInitDisplayMode
		(GLUT_DEPTH | GLUT_RGB | 
		 (NULL != g_arg.ffmpeg ? 
		  GLUT_SINGLE : GLUT_DOUBLE));
	glutInitWindowSize(g_arg.width, g_arg.height);
	glutInitWindowPosition(0, 0);
	glutCreateWindow("Flocks");

	/*
	 * Load textures now into the OpenGL context.
	 * Failure causes no textures to be drawn.
	 */
	if ( ! texLoadAll(&g_arg, fecfg[FECFGKEY_TEXTURES]))
		g_arg.drawnotexs = 1;

	initGL();
	glutMouseFunc(flocks_mouse);
	glutKeyboardFunc(flocks_keyboard);
	glutSpecialFunc(flocks_special);
	glutReshapeFunc(flocks_reshape);
	glutDisplayFunc(flocks_display);
	if (NULL == g_arg.ffmpeg) {
		timerclear(&g_arg.tv);
		glutTimerFunc(g_arg.refresh, flocks_update_timeo, -1);
	} else
		glutIdleFunc(flocks_update_idle);
	g_arg.noisechan = -1;
	if (NULL != g_arg.noise) {
		g_arg.noisechan = Mix_PlayChannel(-1, g_arg.noise, -1);
		assert(-1 != g_arg.noisechan);
		Mix_Volume(g_arg.noisechan, 0);
	}
	glutMainLoop();
	return(EXIT_SUCCESS);
}
