/*	$Id: hud.c,v 1.35 2015/06/06 21:00:42 kristaps Exp $ */
/*
 * Copyright (c) 2014, 2015 Kristaps Dzonsons <kristaps@kcons.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <gd.h>
#include <gsl/gsl_rng.h>
#include <flocks.h>
#include <SDL_mixer.h>

#include "extern.h"

typedef	void (*drawHudNodeFunc)(struct args *);

static	void hudAvgAge(struct args *);
static	void hudAlign(struct args *);
static	void hudAlignclock(struct args *);
static	void hudAlignflock(struct args *);
static	void hudCamera(struct args *);
static	void hudCohesion(struct args *);
static	void hudFlocksz(struct args *);
static	void hudFPS(struct args *);
static	void hudFrames(struct args *);
static	void hudLight0(struct args *);
static	void hudMatchbox(struct args *);
static	void hudMatches(struct args *);
static	void hudPlayInterval(struct args *);
static	void hudScattering(struct args *);
static	void hudSelfbox(struct args *);
static	void hudSeparation(struct args *);
static	void hudSpeedMax(struct args *);
static	void hudSpeedMin(struct args *);
static	void hudTendRdest(struct args *);
static	void hudTendRoost(struct args *);
static	void hudWindow(struct args *);

static	void drawHudText(const char *fmt, ...)
	 	__attribute__((format(printf, 1, 2)));

/*
 * Each "node" (displayable) in the HUD views.
 */
struct	drawHudNode {
	int		 flags;
#define	HUDNODE_EDIT	 1 /* editable field */
#define	HUDNODE_MINI	 2 /* part of mini-HUD */
	drawHudNodeFunc	 fp; /* draw text of HUD field */
	enum cfgkey	 key; /* configuration (if editable) */
	double		 incr; /* increment (if editable) */
};

static	const struct drawHudNode nodes[HUD__MAX] = {
	{ 0, hudAvgAge, CFGKEY__MAX, 0.0 },
	{ HUDNODE_EDIT, hudAlign, CFGKEY_ALIGN, 0.001 },
	{ 0, hudAlignclock, CFGKEY__MAX, 0.0 },
	{ HUDNODE_EDIT, hudAlignflock, CFGKEY_ALIGNFLOCK, 0.01 },
	{ 0, hudCamera, CFGKEY__MAX, 0.0 },
	{ HUDNODE_EDIT, hudCohesion, CFGKEY_COHESION, 0.01 },
	{ HUDNODE_MINI, hudFlocksz, CFGKEY__MAX, 0.0 },
	{ 0, hudFPS, CFGKEY__MAX, 0.0 },
	{ 0, hudFrames, CFGKEY__MAX, 0.0 },
	{ 0, hudLight0, CFGKEY__MAX, 0.0 },
	{ 0, hudMatchbox, CFGKEY__MAX, 0.0 },
	{ 0, hudMatches, CFGKEY__MAX, 0.0 },
	{ HUDNODE_EDIT, hudPlayInterval, CFGKEY_PLAYINTERVAL, 0.1 },
	{ 0, hudScattering, CFGKEY__MAX, 0.0 },
	{ HUDNODE_MINI, hudSelfbox, CFGKEY__MAX, 0.0 },
	{ HUDNODE_EDIT, hudSeparation, CFGKEY_SEPARATION, 0.01 },
	{ HUDNODE_EDIT, hudSpeedMax, CFGKEY_SPEED_MAX, 0.1 },
	{ HUDNODE_EDIT, hudSpeedMin, CFGKEY_SPEED_MIN, 0.1 },
	{ HUDNODE_EDIT, hudTendRdest, CFGKEY_TEND_RDEST, 0.001 },
	{ HUDNODE_EDIT, hudTendRoost, CFGKEY_TEND_ROOST, 0.001 },
	{ HUDNODE_MINI, hudWindow, CFGKEY__MAX, 0.0 },
};

static void
drawHudCharacters(const char *cp)
{

	assert(NULL != cp);
	for ( ; '\0' != *cp; cp++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *cp);
}

static void
drawHudText(const char *fmt, ...)
{
	char	 buf[1024];
	va_list	 ap;

	va_start(ap, fmt);
	(void)vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	drawHudCharacters(buf);
}

static void 
hudAvgAge(struct args *g_arg)
{

	drawHudText("Current Average Age: %.1f s", g_arg->g->ageavg);
}

static void 
hudAlign(struct args *g_arg)
{

	drawHudText("Alignment: %.3f", g_arg->g->align);
}

static void 
hudAlignflock(struct args *g_arg)
{

	drawHudText("Alignment (flock): %.1f%%", 
		100.0 * g_arg->g->alignflock);
}

static void 
hudLight0(struct args *g_arg)
{

	drawHudText("Lighting: %.1f, %.1f, %.1f",
		g_arg->lightpos[0],
		g_arg->lightpos[1],
		g_arg->lightpos[2]);
}

static void 
hudCamera(struct args *g_arg)
{

	drawHudText("Camera: %.1f m facing %.2f, %.2f, %.2f",
		g_arg->cradius, g_arg->cx, g_arg->cy, g_arg->cz);
}

static void
hudAlignclock(struct args *g_arg)
{

	if (g_arg->g->palignt > 0.0)
		drawHudText("Align clock: %.0f s (mean %g)", 
			g_arg->g->palignt,
			g_arg->g->palign);
	else if (g_arg->g->palignflockt > 0.0)
		drawHudText("Align-flock clock: %.0f s (mean %g)", 
			g_arg->g->palignflockt,
			g_arg->g->palignflock);
	else
		drawHudText("Align clock: none");
}

static void
hudScattering(struct args *g_arg)
{

	if (g_arg->g->scattering > 0.0)
		drawHudText("Scattering: %.0f s (mean %g)", 
			g_arg->g->scattering,
			g_arg->g->scattermean);
	else if (g_arg->g->nextscattering > 0.0)
		drawHudText("Scattering: next in %.0f s", 
			g_arg->g->nextscattering);
	else
		drawHudText("Scattering: nope");
}

static void
hudSelfbox(struct args *g_arg)
{
	size_t	 i;

	drawHudText("Lifetime Assortativity:");
	for (i = 0; i < g_arg->g->flocks; i++) {
		if (0 == g_arg->g->selfhitssumsz[i])
			drawHudText(" n/a");
		else
			drawHudText(" %.2f", 
				 g_arg->g->selfhitssum[i] /
				 (double)g_arg->g->selfhitssumsz[i]);
		drawHudText("%s", i < g_arg->g->flocks - 1 ? "," : "");
	}
}

static void
hudMatchbox(struct args *g_arg)
{
	size_t	 i;

	drawHudText("Own-Flock Matches:");
	for (i = 0; i < g_arg->g->flocks; i++)
		drawHudText(" %3.0f%%%s", 
			g_arg->g->matchbox[i][i] / 
			(double)g_arg->g->matchboxhits * 100.0,
			i < g_arg->g->flocks - 1 ? "," : "");
}

static void
hudFPS(struct args *g_arg)
{

	if (g_arg->fps_real < 0.0)
		drawHudText("Frame-Rate: %.2f f/s (want %zu)", 
			g_arg->fps_real, g_arg->fps);
	else
		drawHudText("Frame-Rate: %.1f f/s (want %zu)", 
			g_arg->fps_real, g_arg->fps);
}

static void
hudFrames(struct args *g_arg)
{

	drawHudText("Frames: %zu", g_arg->frame);
}

static void
hudFlocksz(struct args *g_arg)
{
	size_t	 i;

	drawHudText("Population Shares:");
	for (i = 0; i < g_arg->g->flocks; i++)
		drawHudText(" %.2f%s", 
			g_arg->g->flocksz[i] /
			(double)g_arg->g->sz, 
			i < g_arg->g->flocks - 1 ? "," : "");
}

static void 
hudCohesion(struct args *g_arg)
{

	drawHudText("Cohesion: %.2f", g_arg->g->cohesion);
}

static void
hudMatches(struct args *g_arg)
{

	drawHudText("Matches: %zu", g_arg->g->matches);
}

static void 
hudPlayInterval(struct args *g_arg)
{

	drawHudText("Game-Play Interval: %.1f s", g_arg->g->play_interval);
}

static void 
hudSeparation(struct args *g_arg)
{

	drawHudText("Separation: %.2f", g_arg->g->separation);
}

static void 
hudSpeedMax(struct args *g_arg)
{

	drawHudText("Speed (maximum): %.1f m/s", g_arg->g->maxspeed);
}

static void 
hudSpeedMin(struct args *g_arg)
{

	drawHudText("Speed (minimum): %.1f m/s", g_arg->g->minspeed);
}

static void 
hudTendRoost(struct args *g_arg)
{

	drawHudText("Roost Tendency: %.3f", g_arg->g->tend_to_roost);
}

static void 
hudTendRdest(struct args *g_arg)
{

	drawHudText("Removal Edge Tendency: %.3f", g_arg->g->tend_to_rdest);
}

static void
hudWindow(struct args *g_arg)
{
	size_t	 i;

	drawHudText("%zu-Window Assortativity:", 
		g_arg->g->windowsz);
	for (i = 0; i < g_arg->g->flocks; i++) {
		if (0 == g_arg->g->windowhits[i])
			drawHudText(" %s", "n/a");
		else
			drawHudText(" %.2f", 
				g_arg->g->windowmatches[i] / 
				(double)g_arg->g->windowhits[i]);
		drawHudText("%s", i < g_arg->g->flocks - 1 ? "," : "");
	}
}

static void
drawHudHorizSpace(struct args *g_args)
{
	GLint	vals[4];

	glGetIntegerv(GL_CURRENT_RASTER_POSITION, vals);
	glColor4d(1.0, 1.0, 1.0, 1.0);
	glRasterPos2i(vals[0], vals[1]);
	drawHudText(".  ");
}

/*
 * Draw a HUD node on the screen and perform bookkeeping.
 * Namely, we want to note the painted field for intercepting
 * mouse-clicks and routing to the appropriate HUD node for editing.
 * We also want to keep track of our boundaries for painting the
 * background behind the HUD.
 */
static void
drawHudNode(struct args *g_args, enum hudtype type,
	double *minw, double *maxw, double *minh, double *maxh)
{
	GLint	w, h, neww, newh, vals[4];

	/* Get our initial x,y position. */
	glGetIntegerv(GL_CURRENT_RASTER_POSITION, vals);
	g_args->hudnodes[type].minx = w = vals[0];
	g_args->hudnodes[type].miny = h = vals[1];
	g_args->hudnodes[type].maxy = h + 12;

	/* Adjust our initial borders. */
	if (NULL != minw && w < *minw)
		*minw = w;
	if (NULL != minh && h + 12 < *minh)
		*minh = h + 12;
	if (NULL != maxw && w > *maxw)
		*maxw = w;
	if (NULL != maxh && h + 12 > *maxh)
		*maxh = h + 12;

	/*
	 * XXX: for reasons unbeknownst to me, we need to set the colour
	 * before we call glRasterPos.
	 * Switch on whether we're being edited, being selected for
	 * editing, or just an ordinary field.
	 */
	if (1 == g_args->hudnodes[type].edit) 
		glColor4d(0.0, 0.8, 1.0, 1.0);
	else if (2 == g_args->hudnodes[type].edit)
		glColor4d(0.0, 1.0, 1.0, 1.0);
	else 
		glColor4d(1.0, 1.0, 1.0, 1.0);

	glRasterPos2i(w, h);
	(*nodes[type].fp)(g_args);

	/* Get our final x,y position. */
	glGetIntegerv(GL_CURRENT_RASTER_POSITION, vals);
	g_args->hudnodes[type].maxx = neww = vals[0];
	newh = vals[1];

	/* Buffer up or down for the side-HUD views. */
	if (1 == g_args->drawhud) {
		if (HUDNODE_EDIT & nodes[type].flags)
			glRasterPos2i(w, newh += 16);
		else
			glRasterPos2i(w, newh -= 16);
	} 

	/* Readjust our borders. */
	if (NULL != minw && neww < *minw)
		*minw = neww;
	if (NULL != minh && newh < *minh)
		*minh = newh;
	if (NULL != maxw && neww + 12 > *maxw)
		*maxw = neww + 12;
	if (NULL != maxh && newh > *maxh)
		*maxh = newh;

}

/*
 * Intercept the up and down keys for incrementing or decrementing,
 * respectively, a currently-selected editable HUD node.
 */
void
hudSpecial(struct args *arg, int key, int x, int y)
{
	enum hudtype	 sel;
	double		 v;

	assert(arg->drawhud);

	/* Search for the edited HUD node. */
	for (sel = 0; sel < HUD__MAX; sel++)
		if (2 == arg->hudnodes[sel].edit)
			break;

	/* 
	 * Set "v" to be the current value. 
	 * XXX: right now, we only accept double values.
	 * However, there's a cfg_scalar_int() function we can use for
	 * setting non-double values.
	 */
	switch (sel) {
	case (HUD_ALIGN):
		v = arg->g->align;
		break;
	case (HUD_ALIGNFLOCK):
		v = arg->g->alignflock;
		break;
	case (HUD_COHESION):
		v = arg->g->cohesion;
		break;
	case (HUD_PLAY_INTERVAL):
		v = arg->g->play_interval;
		break;
	case (HUD_SEPARATION):
		v = arg->g->separation;
		break;
	case (HUD_SPEED_MAX):
		v = arg->g->maxspeed;
		break;
	case (HUD_SPEED_MIN):
		v = arg->g->minspeed;
		break;
	case (HUD_TEND_RDEST):
		v = arg->g->tend_to_rdest;
		break;
	case (HUD_TEND_ROOST):
		v = arg->g->tend_to_roost;
		break;
	default:
		/* Nothing behind edited. */
		return;
	}

	assert(CFGKEY__MAX != nodes[sel].key);

	if (key == GLUT_KEY_UP)
		cfg_scalar_double(arg->g, 
			nodes[sel].key, v + nodes[sel].incr);
	else if (key == GLUT_KEY_DOWN)
		cfg_scalar_double(arg->g, 
			nodes[sel].key, v - nodes[sel].incr);
}

/*
 * Intercept a mouse click.
 * We look up whether the mouse click occurs fully over an editable HUD
 * field, and if so, to bookkeep that.
 */
void
hudMouse(struct args *arg, int button, int state, int x, int y)
{
	enum hudtype	 i;
	int		 sv;

	for (i = 0; i < HUD__MAX; i++) {
		if ( ! (HUDNODE_EDIT & nodes[i].flags))
			continue;
		/* 
		 * Save then zero the current state. 
		 * This will make exactly one HUD node be editable (or
		 * selected for editing) at a a time.
		 */
		sv = arg->hudnodes[i].edit;
		arg->hudnodes[i].edit = 0;

		if (x < arg->hudnodes[i].minx)
			continue;
		if (x > arg->hudnodes[i].maxx)
			continue;
		if (y < arg->hudnodes[i].miny)
			continue;
		if (y > arg->hudnodes[i].maxy)
			continue;

		/*
		 * If we're pressing down, then indicate that we've
		 * started editing the given field.
		 * If the mouse button is released in the same place,
		 * then mark that the given field is being edited.
		 */
		if (GLUT_DOWN == state)
			arg->hudnodes[i].edit = 1;
		else if (GLUT_UP == state && sv)
			arg->hudnodes[i].edit = 2;
	}
}

/*
 * Draw one of the two "big" HUDs, which are on the left-hand (top and
 * bottom) side of the screen.
 */
static void
hudDrawBig(struct args *g_arg)
{
	double		 minh, maxh, minw, maxw;
	enum hudtype	 h;

	maxh = maxw = -DBL_MAX;
	minh = minw = 0.0;
	glRasterPos2i(12, 12);
	for (h = 0; h < HUD__MAX; h++) {
		if ( ! (HUDNODE_EDIT & nodes[h].flags))
			continue;
		drawHudNode(g_arg, h, &minw, &maxw, &minh, &maxh);
	}

	glBegin(GL_QUADS);
	glColor4d(0.0, 0.0, 0.0, 0.5);
	glVertex2d(minw, minh);
	glVertex2d(minw, maxh);
	glVertex2d(maxw, maxh);
	glVertex2d(maxw, minh);
	glEnd();

	minh = DBL_MAX;
	maxw = -DBL_MAX;
	minw = 0.0;
	maxh = g_arg->height;
	glRasterPos2i(12, g_arg->height - 24);
	for (h = 0; h < HUD__MAX; h++) {
		if (HUDNODE_EDIT & nodes[h].flags)
			continue;
		drawHudNode(g_arg, h, &minw, &maxw, &minh, &maxh);
	}

	glBegin(GL_QUADS);
	glColor4d(0.0, 0.0, 0.0, 0.5);
	glVertex2d(minw, minh);
	glVertex2d(minw, maxh);
	glVertex2d(maxw, maxh);
	glVertex2d(maxw, minh);
	glEnd();
}

/*
 * The mini-HUD just displays pertinent information below the playing
 * field.
 */
static void
hudDrawMini(struct args *g_arg)
{
	enum hudtype	 h;

	glRasterPos2i(12, 12);
	for (h = 0; h < HUD__MAX; h++) {
		if ( ! (HUDNODE_MINI & nodes[h].flags)) 
			continue;
		drawHudNode(g_arg, h, NULL, NULL, NULL, NULL);
		drawHudHorizSpace(g_arg);
	}

	glBegin(GL_QUADS);
	glColor4d(0.0, 0.0, 0.0, 0.5);
	glVertex2d(0.0, 36.0);
	glVertex2d(0.0, 0.0);
	glVertex2d(g_arg->width, 0.0);
	glVertex2d(g_arg->width, 36.0);
	glEnd();
}

static void
hudDrawLastplayer(struct args *g_arg, size_t pos)
{

	drawHudText("Player %zu: type %zu (%.3f %s %.2f", pos,
		g_arg->g->lasttypes[pos],
		g_arg->g->lastconds[pos],
		g_arg->g->lastconds[pos] < 
		 g_arg->g->lastpoffs[pos] ? "<" : ">=",
		g_arg->g->lastpoffs[pos]);
	if (g_arg->g->lastconds[pos] < g_arg->g->lastpoffs[pos])
		drawHudText(" replacing type %zu", g_arg->g->lastkills[pos]);
	drawHudText(")");
}

static void
hudDrawLastplay(struct args *g_arg)
{

	glColor4d(1.0, 1.0, 1.0, 1.0);

	if (0 == g_arg->g->matches) {
		glRasterPos2i(12, g_arg->height - 18);
		drawHudText("Waiting for first play...");
	} else {
		glRasterPos2i(12, g_arg->height - 14);
		hudDrawLastplayer(g_arg, 0);
		glRasterPos2i(12, g_arg->height - 30);
		hudDrawLastplayer(g_arg, 1);
	}

	glBegin(GL_QUADS);
	glColor4d(0.0, 0.0, 0.0, 0.5);
	glVertex2d(0.0, g_arg->height - 38.0);
	glVertex2d(0.0, g_arg->height);
	glVertex2d(g_arg->width, g_arg->height);
	glVertex2d(g_arg->width, g_arg->height - 38.0);
	glEnd();
}

/*
 * Draw our HUD ("head's up display").
 * This is an overlay atop the viewport that shows statistics.
 * In short, it's just text atop a transparent rectangle in orthographic mode.
 * This all comes "in front" of the scene in projection mode.
 */
void
hudDraw(struct args *g_arg)
{

	/* Disable artifacts from the 3D view. */
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND); 
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 

	/* Enter orthographic mode. */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, g_arg->width, 0.0, g_arg->height);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	if (HUDVIEW_LASTPLAY == g_arg->drawhud) {
		hudDrawMini(g_arg);
		hudDrawLastplay(g_arg);
	} else if (HUDVIEW_BIG == g_arg->drawhud) {
		hudDrawBig(g_arg);
	} else if (HUDVIEW_MINI == g_arg->drawhud)
		hudDrawMini(g_arg);

	/* 
	 * Now re-enable the original specification.
	 * Note that we don't reenable projection mode; that is done
	 * whenever the 3d context starts drawing.
	 */
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_BLEND);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
}
