/*	$Id: obj.c,v 1.15 2015/06/05 11:34:57 kristaps Exp $ */
/*
 * Copyright (c) 2014, 2015 Kristaps Dzonsons <kristaps@kcons.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <limits.h>
#include <stdio.h>
#ifdef __linux__
#include <bsd/stdio.h>
#endif
#include <stdlib.h>
#include <string.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <gd.h>
#include <gsl/gsl_rng.h>
#include <flocks.h>
#include <SDL_mixer.h>
#include <zip.h>

#include "extern.h"

struct	obj {
	size_t		 vindex1;
	size_t		 vindex2;
	size_t		 vindex3;
	size_t		 vnindex1;
	size_t		 vnindex2;
	size_t		 vnindex3;
};

static size_t
objLoad(struct zip *arch, const char *fname, float **norm, float **vert)
{
	struct vector	*normals, *vertices;
	struct obj	*faces;
	struct vector	 tmpvert, max, min, sum;
	struct obj	 tmpface;
	size_t		 normalsz, verticesz, facesz, i, j;
	char 		*buf, *line, *end;
	int		 rc, ssz;
	struct zip_file	*file;
	struct zip_stat	 sb;
	int64_t		 idx, sz, rsz;

	zip_stat_init(&sb);

	if (-1 == (idx = zip_name_locate(arch, fname, 0))) {
		fprintf(stderr, "%s: not found\n", fname);
		return(0);
	} else if (zip_stat_index(arch, idx, 0, &sb) < 0) {
		fprintf(stderr, "%s: cannot stat\n", fname);
		return(0);
	} else if (NULL == (file = zip_fopen_index(arch, idx, 0))) {
		fprintf(stderr, "%s: cannot open\n", fname);
		return(0);
	} else if (NULL == (buf = malloc(sb.size + 1))) {
		perror(NULL);
		zip_fclose(file);
		return(0);
	}

	sz = sb.size;
	rsz = 0;
	while (sz > 0) {
		ssz = zip_fread(file, buf + rsz, sz - rsz);
		if (ssz < 0) {
			fprintf(stderr, "%s: cannot read\n", fname);
			zip_fclose(file);
			free(buf);
			return(0);
		} else if (0 == ssz)
			break;
		sz -= ssz;
		rsz += ssz;
	}

	zip_fclose(file);
	buf[rsz] = '\0';

	normals = vertices = NULL;
	faces = NULL;
	normalsz = verticesz = facesz = 0;
	vector_zero(&max);
	vector_zero(&min);
	vector_zero(&sum);

	line = buf;
	for ( ; NULL != (end = strchr(line, '\n')); line = end) {
		*end++ = '\0';

		if ('v' == line[0] && 'n' == line[1])
			rc = 3 == sscanf(line + 2, "%lg %lg %lg", 
				&tmpvert.x, &tmpvert.y, &tmpvert.z);
		else if ('v' == line[0] && ' ' == line[1])
			rc = 3 == sscanf(line + 1, "%lg %lg %lg", 
				&tmpvert.x, &tmpvert.y, &tmpvert.z);
		else if ('f' == line[0] && ' ' == line[1])
			rc = 6 == sscanf(line + 1, 
				"%zu//%zu %zu//%zu %zu//%zu",
				&tmpface.vindex1, &tmpface.vnindex1,
				&tmpface.vindex2, &tmpface.vnindex2,
				&tmpface.vindex3, &tmpface.vnindex3);
		else
			continue;

		if ( ! rc) {
			fprintf(stderr, "%s: bad OBJ\n", fname);
			goto err;
		}

		if ('v' == line[0] && 'n' == line[1]) {
			normals = realloc(normals,
				(normalsz + 1) * sizeof(struct vector));
			normals[normalsz++] = tmpvert;
		} else if ('v' == line[0] && ' ' == line[1]) {
			if (tmpvert.x > max.x)
				max.x = tmpvert.x;
			if (tmpvert.y > max.y)
				max.y = tmpvert.y;
			if (tmpvert.z > max.z)
				max.z = tmpvert.z;
			if (tmpvert.x < min.x)
				min.x = tmpvert.x;
			if (tmpvert.y < min.y)
				min.y = tmpvert.y;
			if (tmpvert.z < min.z)
				min.z = tmpvert.z;
			sum.x += tmpvert.x;
			sum.y += tmpvert.y;
			sum.z += tmpvert.z;
			vertices = realloc(vertices,
				(verticesz + 1) * sizeof(struct vector));
			vertices[verticesz++] = tmpvert;
		} else if ('f' == line[0] && ' ' == line[1]) {
			faces = realloc(faces,
				(facesz + 1) * sizeof(struct obj));
			faces[facesz++] = tmpface;
		}
	}

	sum.x /= (double)verticesz;
	sum.y /= (double)verticesz;
	sum.z /= (double)verticesz;

	vector_zero(&max);
	vector_zero(&min);

	for (i = 0; i < verticesz; i++) {
		vertices[i].x = (vertices[i].x - sum.x);
		if (vertices[i].x > max.x)
			max.x = vertices[i].x;
		if (vertices[i].x < min.x)
			min.x = vertices[i].x;
		vertices[i].y = (vertices[i].y - sum.y);
		if (vertices[i].y > max.y)
			max.y = vertices[i].y;
		if (vertices[i].y < min.y)
			min.y = vertices[i].y;
		vertices[i].z = (vertices[i].z - sum.z);
		if (vertices[i].z > max.z)
			max.z = vertices[i].z;
		if (vertices[i].z < min.z)
			min.z = vertices[i].z;
	}

	*norm = malloc(sizeof(float) * 3 * 3 * facesz);
	*vert = malloc(sizeof(float) * 3 * 3 * facesz);

	for (i = j = 0; i < facesz; i++, j += 9) {
		(*norm)[j + 0] = normals[faces[i].vnindex1 - 1].x;
		(*norm)[j + 1] = normals[faces[i].vnindex1 - 1].y;
		(*norm)[j + 2] = normals[faces[i].vnindex1 - 1].z;
		(*norm)[j + 3] = normals[faces[i].vnindex2 - 1].x;
		(*norm)[j + 4] = normals[faces[i].vnindex2 - 1].y;
		(*norm)[j + 5] = normals[faces[i].vnindex2 - 1].z;
		(*norm)[j + 6] = normals[faces[i].vnindex3 - 1].x;
		(*norm)[j + 7] = normals[faces[i].vnindex3 - 1].y;
		(*norm)[j + 8] = normals[faces[i].vnindex3 - 1].z;

		(*vert)[j + 0] = vertices[faces[i].vindex1 - 1].x;
		(*vert)[j + 1] = vertices[faces[i].vindex1 - 1].y;
		(*vert)[j + 2] = vertices[faces[i].vindex1 - 1].z;
		(*vert)[j + 3] = vertices[faces[i].vindex2 - 1].x;
		(*vert)[j + 4] = vertices[faces[i].vindex2 - 1].y;
		(*vert)[j + 5] = vertices[faces[i].vindex2 - 1].z;
		(*vert)[j + 6] = vertices[faces[i].vindex3 - 1].x;
		(*vert)[j + 7] = vertices[faces[i].vindex3 - 1].y;
		(*vert)[j + 8] = vertices[faces[i].vindex3 - 1].z;
	}

	free(buf);
	free(faces);
	free(vertices);
	free(normals);
	return(facesz * 3);
err:
	free(buf);
	free(vertices);
	free(faces);
	free(normals);
	return(0);
}

int
objLoadAll(struct args *arg, const char *fname)
{
	struct zip	*arch;
	size_t		 i;
	char		 buf[PATH_MAX];

	arch = zip_open(NULL == fname ?
		DATADIR "/objects.zip" : fname, 0, NULL);

	if (NULL == arch) {
		fprintf(stderr, "%s: not a ZIP\n", fname);
		return(0);
	}

	for (i = 0; i < WING_SIZE; i++) {
		snprintf(buf, sizeof(buf), "bird.%zu.obj", i);
		arg->birdNumVerts[i] = objLoad
			(arch, buf, &arg->birdNormals[i], 
			 &arg->birdVerts[i]);
		if (0 == arg->birdNumVerts[i]) {
			zip_discard(arch);
			return(0);
		}
	}

	zip_discard(arch);
	return(i == WING_SIZE);
}
