/*	$Id: tex.c,v 1.10 2015/06/05 11:34:57 kristaps Exp $ */
/*
 * Copyright (c) 2014, 2015 Kristaps Dzonsons <kristaps@kcons.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <gd.h>
#include <gsl/gsl_rng.h>
#include <flocks.h>
#include <SDL_mixer.h>
#include <zip.h>

#include "extern.h"

static	const char *faces[FACE__MAX] = {
	"posx.jpg", /* FACE_POSX */
	"posy.jpg", /* FACE_POSY */
	"posz.jpg", /* FACE_POSZ */
	"negx.jpg", /* FACE_NEGX */
	"negy.jpg", /* FACE_NEGY */
	"negz.jpg", /* FACE_NEGZ */
};

static int
tex_getc(struct gdIOCtx *ctx)
{
	int	 c;

	if (zip_fread(ctx->data, &c, 1) <= 0)
		return(EOF);
	return(c);
}

static int
tex_getbuf(struct gdIOCtx *ctx, void *buf, int sz)
{
	int	 ret;

	if (-1 == (ret = zip_fread(ctx->data, buf, sz)))
		return(EOF);
	return(ret);
}

static unsigned char *
texLoad(struct zip *arch, const char *fname, size_t *w, size_t *h)
{
	gdImagePtr	 im;
	size_t		 i, j, k, sw, sh;
	int		 c;
	unsigned char	*data;
	int64_t		 index;
	struct zip_file	*f;
	gdIOCtx		 ctx;

	if (-1 == (index = zip_name_locate(arch, fname, 0))) {
		fprintf(stderr, "%s: not found in ZIP\n", fname);
		return(NULL);
	} else if (NULL == (f = zip_fopen_index(arch, index, 0))) {
		fprintf(stderr, "%s: couldn't open\n", fname);
		return(NULL);
	}

	memset(&ctx, 0, sizeof(struct gdIOCtx));
	ctx.getC = tex_getc;
	ctx.getBuf = tex_getbuf;
	ctx.data = f;

	if (NULL == (im = gdImageCreateFromJpegCtx(&ctx))) {
		fprintf(stderr, "%s: couldn't read JPG\n", fname);
		zip_fclose(f);
		return(NULL);
	}
	zip_fclose(f);

	*w = sw = gdImageSX(im);
	*h = sh = gdImageSY(im);

	data = malloc(sw * sh * 3);
	assert(NULL != data);
	
	for (k = i = 0; i < sh; i++) {
		for (j = 0; j < sw; j++) {
			c = gdImageGetPixel(im, (int)j, (int)(sh - i - 1));
			data[k++] = gdImageRed(im, c);
			data[k++] = gdImageGreen(im, c);
			data[k++] = gdImageBlue(im, c);
		}
	}

	gdImageDestroy(im);
	return(data);
}

int
texLoadAll(struct args *args, const char *fname)
{
	enum face	 face;
	size_t		 w, h;
	unsigned char	*dat;
	struct zip	*arch;

	arch = zip_open(NULL == fname ?
		DATADIR "/textures.zip" : fname, 0 , NULL);

	if (NULL == arch) {
		fprintf(stderr, "%s: not a ZIP\n", fname);
		return(0);
	}

	for (face = 0; face < FACE__MAX; face++) {
		if (NULL == (dat = texLoad(arch, faces[face], &w, &h))) 
			break;
		glGenTextures(1, &args->bgid[face]);
		glBindTexture(GL_TEXTURE_2D, args->bgid[face]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
			w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, dat);
		glTexParameteri(GL_TEXTURE_2D, 
			GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, 
			GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, 
			GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, 
			GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		free(dat);
	}

	zip_discard(arch);
	return(face == FACE__MAX);
}
